#Code 9.8 Phenomenological model extracted from single observed variable (pheno_1.R)

rm(list=ls(all=TRUE))  #remove objects between runs
options(scipen=999)  #disable scientific notation

#Directories
dir.udf<-as.character("~/Documents/BOLOGNA/Didattica/NLTS/pheno/UserDefinedFunctions")
dir.data<-as.character("~/Documents/BOLOGNA/Didattica/NLTS/pheno/data")
dir.results<-as.character("~/Documents/BOLOGNA/Didattica/NLTS/pheno/results")

#Load user defined functions
setwd(dir.udf)
dump("lorenz_udf", file="lorenz_udf.R")
source("lorenz_udf.R")
dump("embed_udf", file="embed_udf.R")
source("embed_udf.R")
dump("finite_differences_udf", file="finite_differences_udf.R")
source("finite_differences_udf.R")
dump("polynomial_udf", file="polynomial_udf.R")
source("polynomial_udf.R")
dump("ols_udf", file="ols_udf.R")
source("ols_udf.R")
dump("lasso_udf", file="lasso_udf.R")
source("lasso_udf.R")
dump("goodfit_3_udf", file="goodfit_3_udf.R")
source("goodfit_3_udf.R")
dump("solve_ode_3_udf", file="solve_ode_3_udf.R")
source("solve_ode_3_udf.R")
dump("spectral_udf", file="spectral_udf.R")
source("spectral_udf.R")

#Generate Lorenz data with user-defined function 'lorenz_udf'
lorenz_results<-lorenz_udf(x0=1,y0=1,z0=1,sigma=10,beta=8/3,rho=28,
 t.end=50,delta=0.001)
x.ts<-lorenz_results[[1]]

#Set parameters
delta<-0.01  #integration step
m<-3  #number of equations
Mp<-3  #polynomial range
alpha<-0.5  #glmnet elastic-net mixing parameter
weight<-FALSE  #run adaptive LASSO
nu<-1  #Adaptive LASSO non-negative tuning parameter
ci<-FALSE  #95% confidence intervals for NSE
method<-"lsoda"
method_spec<-TRUE

#Synthesize dynamic variables by embedding observed time series
results.embed<-embed_udf(x.ts)  #use to compute embedding dimension
d<-results.embed[[1]]
library(tseriesChaos)
Mx<-embedd(x.ts,m=3,d) #embedded data matrix whose columns are x,y,z
x.emb<-Mx[,1];y.emb<-Mx[,2];z.emb<-Mx[,3] #Note that x,y,z are shorter than x.ts
library(scatterplot3d)
scatterplot3d(x.emb,y.emb,z.emb,type='l',lwd=1,cex.axis=1.1,
  cex.lab=1.5,font.axis=2,font.lab=2)

#Response Variable:  User-defined function to compute finite differences
diff_x<-diff(x.emb,delta)  #run user-defined function for x
diff_y<-diff(y.emb,delta)
diff_z<-diff(z.emb,delta)
x<-diff_x[[1]];y<-diff_y[[1]];z<-diff_z[[1]]  #times series same length as differences
x.diff<-diff_x[[2]];y.diff<-diff_y[[2]];z.diff<-diff_z[[2]] #finite differences

#Regressors: User-defined function to compute multivariate polynomial expansion
poly.exp<-poly(m,Mp)
#Data Check: compare order of terms in 'mpoly' with column names
termList2<-poly.exp[[1]]  #nomomial terms
poly1<-mpoly(termList2);term<-terms(poly1);term
col.names<-poly.exp[[2]];col.names
#Monomials populated with data
data1<-poly.exp[[3]]  #populated data matrix
data2<-data1[complete.cases(data1),]  #remove last row with na's
data<-data2[,-1] #remove 1st column of 1's (regression packages don't require))

#User-defined function to run OLS
results_ols_x<-ols_udf(x.diff,data,m,Mp)  #x.diff
xp.ols<-results_ols_x[[1]]
table_ols_x<-results_ols_x[[2]]

results_ols_y<-ols_udf(y.diff,data,m,Mp)  #y.diff
yp.ols<-results_ols_y[[1]]
table_ols_y<-results_ols_y[[2]]

results_ols_z<-ols_udf(z.diff,data,m,Mp)  #z.diff
zp.ols<-results_ols_z[[1]]
table_ols_z<-results_ols_z[[2]]

#User-defined function to run LASSO
results_lasso_x<-lasso_udf(data,x.diff,alpha,weight,nu)
xp.min<-as.vector(results_lasso_x[[1]][,1])
xp.1se<-as.vector(results_lasso_x[[1]][,2])
dev_x<-results_lasso_x[[2]]

results_lasso_y<-lasso_udf(data,y.diff,alpha,weight,nu)
yp.min<-as.vector(results_lasso_y[[1]][,1])
yp.1se<-as.vector(results_lasso_y[[1]][,2])
dev_y<-results_lasso_y[[2]]

results_lasso_z<-lasso_udf(data,z.diff,alpha,weight,nu)
zp.min<-as.vector(results_lasso_z[[1]][,1])
zp.1se<-as.vector(results_lasso_z[[1]][,2])
dev_z<-results_lasso_z[[2]]

#Estimated coefficients to use in solution of ODE system
#xp<-xp.ols;yp<-yp.ols;zp<-zp.ols
xp<-xp.min;yp<-yp.min;zp<-zp.min
#xp<-xp.1se;yp<-yp.1se;zp<-zp.1se

#User-defined function to assess goodness of fit
results_gf<-goodfit_3(x.diff,y.diff,z.diff,x,y,z,xp,yp,zp,Mp,ci)
dx<-results_gf[[1]];dy<-results_gf[[2]];dz<-results_gf[[3]]
fit_table<-results_gf[[4]]

#User-defined function to solve phenomenological model
results_solve<-solve_3(x,y,z,xp,yp,zp,Mp,delta,method)

#Solutions
out<-results_solve[[1]]
plot(out) #plot variables
x.sim<-out[,"x"];length(x.sim)<-length(x)  
y.sim<-out[,"y"];length(y.sim)<-length(y)
z.sim<-out[,"z"];length(z.sim)<-length(z)

#Spectral Analysis of ODE model solution for x (x.sim)
#methods="pgram", "ar"
x.spec<-spectral(x.sim,method_spec=TRUE)
cycles.x<-x.spec[[1]]
y.spec<-spectral(x.sim,method_spec=TRUE)
cycles.y<-y.spec[[1]]
z.spec<-spectral(x.sim,method_spec=TRUE)
cycles.z<-z.spec[[1]]
spec.table<-cbind(cycles.x,cycles.y,cycles.z)

#Plots
#Computed and simulated differentials
par(mfrow=c(4,1)) 
yrange<-range(dx,x.diff)
plot(dx,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(x.diff,lty=1,lwd=2)
yrange<-range(dy,y.diff)
plot(dy,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(y.diff,lty=1,lwd=2)
yrange<-range(dz,z.diff)
plot(dz,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(z.diff,lty=1,lwd=2)

#Observed time series and isolated signal
par(mfrow=c(4,1))
yrange<-range(x.obs,x)
plot(x.obs,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(x,lty=1,lwd=2)
yrange<-range(y.obs,y)
plot(y.obs,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(y,lty=1,lwd=2)
yrange<-range(z.obs,z)
plot(z.obs,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(z,lty=1,lwd=2)

#Signal and simulated signal
par(mfrow=c(4,1))
yrange<-range(x,x.sim)
plot(x,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(x.sim,lty=1,lwd=2)
yrange<-range(y,y.sim)
plot(y,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(y.sim,lty=1,lwd=2)
yrange<-range(z,z.sim)
plot(z,type='l',lwd=4,cex.lab=1.5,cex.axis=1,font=2,font.lab=2,col="dark grey")
lines(z.sim,lty=1,lwd=2)

#Reconstructed attractor
library(scatterplot3d)
scatterplot3d(x,y,z,type='l',lwd=1,cex.lab=1.5,cex.axis=1,
  lab=c(2,2),lab.z=2,font=2,font.lab=2)

#Simulated attractor
scatterplot3d(x.sim,y.sim,z.sim,type="l",lwd=1,cex.lab=1.5,cex.axis=1,
  lab=c(2,2),lab.z=2,font=2,font.lab=2) 

#Export Output
setwd(dir.results)

#Table with coefficients from all regressions
coeffs_all<-cbind(col.names,xp.ols,yp.ols,zp.ols,xp.min,yp.min,
  zp.min,xp.1se,yp.1se,zp.1se)
colnames(coeffs_all)<-c("terms","xp.ols","yp.ols","zp.ols",
  "xp.min","yp.min","zp.min","xp.1se","yp.1se","zp.1se")
write.table(coeffs_all,sep=",","coeffs_all.csv",row.names=FALSE)

#Goodness of fit
write.csv(fit_table,"fit_table.csv")

#Simulated values
sim.val<-cbind(x.sim,y.sim,z.sim)
write.table(sim.val,sep=",",col.names=c("x.sim","y.sim","z.sim"),
            "sim.val.csv",row.names=FALSE)

#Spectal results
write.table(spec.table,sep=",",col.names=c("cycles.x","cycles.y","cycles.z"),
  "cycle.length.csv",row.names=FALSE)